package com.baidu.cloud.videoplayer.demo.helper;

import android.content.Context;
import android.util.Log;

import com.baidu.cloud.media.player.feature.FeatureManager;
import com.baidu.cloud.videoplayer.demo.BaseApp;
import com.baidu.cloud.videoplayer.demo.C;
import com.baidu.cloud.videoplayer.demo.info.SharedPrefsStore;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;

public class CommonUtils {
    private static final String TAG = "CommonUtils";

    public static boolean getSystemProperty(String key, boolean def) {
        try {
            final ClassLoader cl = ClassLoader.getSystemClassLoader();
            final Class<?> systemProperties = cl.loadClass("android.os.SystemProperties");
            final Class<?>[] paramTypes = new Class[] { String.class, Boolean.TYPE };
            final Method get = systemProperties.getMethod("getBoolean", paramTypes);
            final Object[] params = new Object[] { key, def };
            return (boolean) get.invoke(systemProperties, params);
        } catch (Exception e) {
            e.printStackTrace();
            return def;
        }
    }

    /**
     * Find specific class field
     * @param obj :    class object
     * @param fName : field name
     * @return : specific field
     */
    public static Field getDeclaredField(Object obj, String fName){
        Field field = null;

        if (obj == null) {
            return null;
        }
        try {
            Class cls = obj.getClass();

            /* Loop to upper class, and get specific member field */
            for (; cls != Object.class; cls = cls.getSuperclass()) {
                try {
                    field = cls.getDeclaredField(fName);
                    break;
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if (field != null) {
                /* If it is private authority, modify access limit */
                if (!field.isAccessible()) field.setAccessible(true);
            }
        } catch (Exception e) {
            Log.e(TAG, "get " + fName + " field failed: " + e);
        } finally {
            if (field == null) {
                Log.e(TAG, "not found " + fName + " field");
            }
        }
        return field;
    }

    /**
     * Get specific class field value
     * @param obj :    class object
     * @param field : field object
     * @return : specific field value
     */
    public static Object getFieldValue(Object obj, Field field) {
        Object retObject = null;

        if (field != null) {
            try {
                retObject = field.get(obj);
            } catch (Exception e) {
                Log.e(TAG, "get field failed: " + e);
            }
        } else if (obj != null) {
            Log.e(TAG, obj.getClass() + " field is null");
        }
        return retObject;
    }

    /**
     * Invoke static class member method, find method, and call method
     * @param cName : class name
     * @param mName : method name
     * @param types : type array of parameters
     * @param params: value array of parameters
     * @return : method return value
     */
    public static Object invokeStaticMethod(
            String cName, String mName, Class[] types, Object[] params) {
        Object retObject = null;

        try {
            Class<?> cls = Class.forName(cName);
            Method meth = null;

            /* Loop to upper class, and get specific member method */
            for (; cls != Object.class; cls = cls.getSuperclass()) {
                try {
                    meth = cls.getMethod(mName, types);
                    break;
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if (meth != null) {
                /* If it is private authority, modify access limit */
                if (!meth.isAccessible()) meth.setAccessible(true);
                retObject = meth.invoke(null, params);
            } else {
                Log.e(TAG, "not found " + mName + " method");
            }
        } catch (Exception e) {
            Log.e(TAG, "invoke " + mName + " method failed: " + e);
        }
        return retObject;
    }

    /**
     * New class instance with params constructor method
     * @param cName : class name
     * @param types : type array of parameters
     * @param params: value array of parameters
     * @return : class object
     */
    public static Object Load(String cName, Class[] types, Object[] params) {
        Object retObject = null;
        try {
            Class<?> cls = Class.forName(cName);

            Constructor ct = cls.getDeclaredConstructor(types);
            /* If it is private authority, modify access limit */
            if (!ct.isAccessible()) {
                ct.setAccessible(true);
            }
            retObject = ct.newInstance(params);
        } catch (Exception e) {
            Log.e(TAG, "new " + cName + " class failed: " + e);
            e.printStackTrace();
        }
        return retObject;
    }

    /**
     * Invoke class member method,disregard private,protected and default
     * @param obj : class object
     * @param mName : method name
     * @param types : type array of parameters
     * @param params: value array of parameters
     * @return : method return value
     */
    public static Object Load(
            Object obj, String mName, Class[] types, Object[] params) {
        Object retObject = null;

        if(obj == null) return null;
        try {
            Class<?> cls = obj.getClass();
            Method meth = null;

            /* Loop to upper class, and get specific member method */
            for(; cls != Object.class; cls = cls.getSuperclass()) {
                try {
                    meth = cls.getDeclaredMethod(mName, types);
                    break;
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
            if (meth != null) {
                /* If it is private authority, modify access limit */
                if (!meth.isAccessible()) meth.setAccessible(true);
                retObject = meth.invoke(obj, params);
            } else {
                Log.e(TAG, "not found " + mName + " method");
            }
        } catch (Exception e) {
            Log.e(TAG, "invoke " + mName  + " method failed: " + e);
        }
        return retObject;
    }

    public static int getSmartVisionType(Context context) {
        int type = SharedPrefsStore.getIntData(context, C.KEY_SMARTVISION_TYPE, C.SMARTVISION_DISABLE);
        try {
            if (type == C.SMARTVISION_HDR
                    && FeatureManager.getInstance().checkHdrFeatureValid(context, BaseApp.APP_ID)) {
                return C.SMARTVISION_HDR;
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        return C.SMARTVISION_DISABLE;
    }
}
